#!/usr/bin/env python3
"""
bamprep.py - prepare SAM/BAM file for phase change detection

this script will automatically try to detect your samtools binary by default -
the one it uses can be found by running `which samtools` in the command line
(alternatively, you could symlink your preferred version of samtools into a bin/ folder
in your project dir to keep paths clean, and then provide `./bin/samtools`
for the path - recommended)
"""

import os
import re
import sys
import argparse
import subprocess

class ReadcombParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f'error: {message}\n')
        self.print_help()
        sys.exit(1)

def arg_parser():
    parser = ReadcombParser(
        description='prep SAM/BAM for phase change detection',
        usage='readcomb-bamprep [options]')

    parser.add_argument('-b', '--bam', required=True,
        type=str, help='BAM file to prepare')
    parser.add_argument('-s', '--samtools', required=False,
        type=str, help='Path to samtools binary [optional]')
    parser.add_argument('-t', '--threads', required=False,
        type=int, help='Number of threads to use [optional]')
    parser.add_argument('--index_csi', required=False,
        action='store_true', help='Use CSI index instead of BAI [optional]')
    parser.add_argument('--no_progress', required=False,
        action = 'store_true', help='Skip positional sorting (faster, but disables progress \
        bars in filtering step).')
    parser.add_argument('-o', '--outdir', required=False,
        type=str, help='Directory to write to [optional] (default: current dir)')
    parser.add_argument('--version', action='version', version='readcomb 0.3.13')

    return parser

def bamprep(args):
    """
    Prepare BAM for phase change detection with readcomb.

    Script will:
    1. Apply filters (proper pair, not secondary, not supplementary)
    2. Sort by read name instead of position (samtools sort)
    3. Create .bai file (samtools index)

    The output file will be named similarly to the input - only
    the extension will have been changed.

    Parameters
    ----------
    args : Namespace
        Namespace containing all user given arguements compiled by arg_parse()
    """
    # check extension
    match = re.search(r'\.[sb]am$', args.bam)
    if match:
        extension = match.group()
        basename = os.path.basename(args.bam[:args.bam.find(extension)])
    else:
        raise ValueError('input file is not SAM or BAM - make sure extension is correct')

    # get samtools path if not specified
    if not args.samtools:
        samtools_check_cmd = 'which samtools'
        proc = subprocess.Popen(samtools_check_cmd.split(' '),
                stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        out, _ = proc.communicate()
        if not out:
            raise RuntimeError("""ERROR: samtools not found. Please
                    either ensure samtools is in your PATH or specify a
                    samtools binary using the --samtools argument""")
        else:
            args.samtools = out.decode('utf-8').rstrip('\n')

    # format args
    if args.outdir:
        if not args.outdir.endswith('/'):
            args.outdir += '/'
    elif not args.outdir:
        args.outdir = './'
        print('[readcomb] no outdir specified, saving to current working directory')

    # filter out non proper pair, supplementary, and secondary reads
    print('[readcomb] Filtering...')
    filter_pos_cmd = f'{args.samtools} view -O bam '
    filter_pos_cmd += '-f 0x2 -F 0x100 -F 0x800 '
    filter_pos_cmd += f'-o {args.outdir}{basename}.temp.filtered.bam {args.bam}'
    print(f'[readcomb] cmd: {filter_pos_cmd}')
    proc = subprocess.run(filter_pos_cmd.split(' '), check=True)

    if not args.no_progress:
    # sort positionally for bai file creation
        sort_pos_cmd = f'{args.samtools} sort -O bam '
        if args.threads:
            sort_pos_cmd += f'-@{args.threads} '
        sort_pos_cmd += f'-o {args.outdir}{basename}.sorted.bam '
        sort_pos_cmd += f'{args.outdir}{basename}.temp.filtered.bam'
        print('[readcomb] Sorting positionally for .bai creation...')
        print(f'[readcomb] cmd: {sort_pos_cmd}')
        proc = subprocess.run(sort_pos_cmd.split(' '), check=True)

        # create bai file
        if args.index_csi:
            idx_type = '.csi'
        else:
            idx_type = '.bai'
        print(f'[readcomb] Creating {idx_type} index file...')
        index_cmd = f'{args.samtools} index {args.outdir}{basename}.sorted.bam '
        index_cmd += f'{args.outdir}{basename}.sorted{idx_type}'
        print(f'[readcomb] cmd: {index_cmd}')
        proc = subprocess.run(index_cmd.split(' '), check=True)

        # remove positionally sorted file
        try:
            os.remove(f'{args.outdir}{basename}.sorted.bam')
        except OSError as e:
            raise FileNotFoundError(f'Unable to delete {e.filename} - {e.strerror}') from e

    # sort by name and ensure file is in BAM format
    print('[readcomb] Creating read name sorted file...')
    sort_cmd = f'{args.samtools} sort -n -O bam '
    if args.threads:
        sort_cmd += f'-@{args.threads} '
    sort_cmd += f'-o {args.outdir}{basename}.sorted.bam {args.outdir}{basename}.temp.filtered.bam'
    print(f'[readcomb] cmd: {sort_cmd}')
    proc = subprocess.run(sort_cmd.split(' '), check=True)

    # remove filtered unsorted file
    print('[readcomb] Removing temp files...')
    try:
        os.remove(f'{args.outdir}{basename}.temp.filtered.bam')
    except OSError as e:
        raise FileNotFoundError(f'Unable to delete {e.filename} - {e.strerror}')

def main():
    parser = arg_parser()
    # Namespace of arguments
    args = parser.parse_args()

    if args.samtools:
        if not os.path.isfile(args.samtools):
            raise ValueError(f'ERROR: No samtools binary found at {args.samtools}')
    bamprep(args)
    print('[readcomb] Complete.')


if __name__ == '__main__':
    main()



"""
Script to filter out false positives caused by misalignment and/or
sketchy regions of the genome

Three options:
1. Midpoint - creates/uses bed file of false positive phase changes to compare
against
2. Overlap - compares all variants shared by false positive read and putative
recombinant read
3. Nuclear - removes any read that overlaps a false positive read at all

"""

import os
import sys
import csv
import argparse
import ast
import subprocess
import datetime

import pysam
from tqdm import tqdm

try:
    from readcomb.classification import pairs_creation
    from readcomb.filter import ReadcombParser
except ImportError as e:
    print('WARNING: readcomb is not installed')
    print('Use command: pip install readcomb')
    from classification import pairs_creation
    from filter import ReadcombParser 

def arg_parser():
    parser = ReadcombParser(
        description='filter putative rcmb events for alignment-derived false positives', 
        usage='readcomb-fp [options]')

    parser.add_argument('-f', '--fname', required=True, type=str, 
        help='Cross SAM output by readcomb-filter')
    parser.add_argument('-fp', '--false_plus', required=False, type=str, 
        help='False positives from plus parent')
    parser.add_argument('-fm', '--false_minus', required=False, type=str, 
        help='False positives from minus parent')
    parser.add_argument('-v', '--vcf', required=True, type=str, 
        help='VCF with calls for cross of interest')
    parser.add_argument('-m', '--method', required=True, type=str, 
        help='Filtering method to use (midpoint/overlap/nuclear) - see docs')
    parser.add_argument('--false_bed_out', required=False, type=str, 
        help='Where to save false phase change bed file [optional]')
    parser.add_argument('--false_bed_in', required=False, type=str, 
        help='Bed file of false phase changes to use, if already generated [optional]')
    parser.add_argument('--log', required=True, type=str, default='fp.log',
        help='Path to log file')
    parser.add_argument('-o', '--out', required=True, type=str,
        help='File to write filtered reads to')
    parser.add_argument('--version', action='version', version='readcomb 0.3.13')

    return parser

class BedGenerator():
    def __init__(self, false_plus, false_minus, vcf_fname, method, bed_fname):
        """Generate BED file of false positive midpoints

        BED format will depend on what filtering method is requested -

        - ``midpoint`` filtering will generate rows where positions correspond to the two
          variants that denote a phase change
        - ``overlap`` filtering will generate rows that also contain the output of
          `classification.Pair.detection()` - lists of tuples denoting sites and haps
        - ``nuclear`` filtering will only include positions of reads - overlapping
          reads will have their reads reduced

        Parameters
        ----------
        false_plus : str
            path to readcomb-filter output SAM file when run on parent 1
        false_minus : str
            path to readcomb-filter output SAM file when run on parent 2
        vcf_fname : str
            path to VCF used to call recombination events
        method : str
            one of 'midpoint', 'overlap', or 'nuclear'
        bed_fname : str
            path to bed outfile
        """
        # paths to the plus and minus false positive files
        self.false_plus = false_plus
        self.false_minus = false_minus
        self.vcf_fname = vcf_fname

        self.method = method
        self.bed_fname = bed_fname
        if not self.bed_fname: # if fname not provided - generate temp bed
            self.temp = True
            self.bed_fname = 'readcomb.parental_fp.bed.temp'
        
    def generate_bed(self):
        """main method to invoke specific BED creation methods
        """
        if self.method == 'midpoint':
            self._generate_midpoint_bed()
        elif self.method == 'overlap':
            self._generate_overlap_bed()
        elif self.method == 'nuclear':
            self._generate_nuclear_bed()
        else:
            raise ValueError(
                f'method must be one of midpoint, overlap, or nuclear, not {self.method}')

    def _generate_midpoint_bed(self):
        """generate BED file for midpoint method

        the format is as follows:
        #chrom	start	end	hap1	hap2	parental_file

        where the start and end columns correspond to the two closest variants
        surrounding a phase change, and parental_file denotes whether this
        false positive was in the plus or minus sequence
        """
        print('[readcomb] generating bed for midpoint method filtering...')
        plus_reader = pairs_creation(self.false_plus, self.vcf_fname)
        minus_reader = pairs_creation(self.false_minus, self.vcf_fname)

        f = open(self.bed_fname, 'w')
        fieldnames = ['#chrom', 'start', 'end', 'hap1', 'hap2', 'parental_file']
        bed_writer = csv.DictWriter(f, delimiter='\t', fieldnames=fieldnames)
        bed_writer.writeheader()

        # iterate through false positives
        for i, reader in enumerate([plus_reader, minus_reader]):
            for pair in tqdm(reader, desc=f'parental false positive bed generation {i+1}'):
                pair.classify(masking=0)
                if pair.call == 'no_phase_change': # ignore if no phase change
                    continue
                else:
                    detection_filt = [
                        variant for variant in pair.detection 
                        if variant[0] != 'N']
                    prev_hap, prev_pos = detection_filt[0][:2] # set to first hap
                    for variant in detection_filt:
                        hap, pos, base = variant
                        if hap == prev_hap:
                            prev_pos = pos # update variable
                            continue
                        elif hap != prev_hap:
                            bed_writer.writerow({
                                '#chrom': pair.rec_1.reference_name, 
                                'start': prev_pos, 'end': pos, 
                                'hap1': prev_hap, 'hap2': hap, 
                                'parental_file': 'plus' if i == 0 else 'minus'}
                                )
                            prev_hap = hap
                            prev_pos = pos
        f.close()
    

    def _generate_overlap_bed(self):
        """generate BED file for overlap method

        the format is as follows:
        #chrom	start	end	parental_file	call	detection

        where the detection column contains the literal output of
        ``Pair.detection``, for comparison with putative recombinant pairs of
        interest in full, while call contains the recombination event the false
        positive was called as
        """
        print('[readcomb] generating bed for overlap method filtering...')
        plus_reader = pairs_creation(self.false_plus, self.vcf_fname)
        minus_reader = pairs_creation(self.false_minus, self.vcf_fname)

        f = open(self.bed_fname, 'w')
        fieldnames = ['#chrom', 'start', 'end', 'parental_file', 'call', 'detection']
        bed_writer = csv.DictWriter(f, delimiter='\t', fieldnames=fieldnames)
        bed_writer.writeheader()

        # iterate through false positives
        for i, reader in enumerate([plus_reader, minus_reader]):
            for pair in tqdm(reader, desc=f'parental false positive bed generation {i+1}/2'):
                pair.classify(masking=0)
                if pair.call == 'no_phase_change': # ignore if no phase change
                    continue
                else:
                    bed_writer.writerow({
                        '#chrom': pair.rec_1.reference_name,
                        'start': pair.condensed[0][1],
                        'end': pair.condensed[-1][2],
                        'parental_file': 'plus' if i == 0 else 'minus',
                        'call': pair.call, 'detection': pair.detection}
                        )
        f.close()


    def _generate_nuclear_bed(self):
        """generate bed file for nuclear method

        the format is as follows:
        #chrom	start	end

        this method just stores intervals overlapped by false positive reads in
        either parent - which means the bed first needs to be generated, and
        then 'reduced'

        the reduction is done by calling on self.sort_tabix_bed without
        actually running tabix, and using the temp sorted file to reduce
        overlapping intervals. the final output file is then named the original
        filename provided to self.bed_fname
        """
        print('[readcomb] generating bed for nuclear method filtering...')
        plus_reader = pairs_creation(self.false_plus, self.vcf_fname)
        minus_reader = pairs_creation(self.false_minus, self.vcf_fname)

        f = open(self.bed_fname, 'w')
        fieldnames = ['#chrom', 'start', 'end']
        bed_writer = csv.DictWriter(f, delimiter='\t', fieldnames=fieldnames)
        bed_writer.writeheader()
        
        # iterate through false positives
        for i, reader in enumerate([plus_reader, minus_reader]):
            for pair in tqdm(reader, desc=f'parental false positive bed generation {i+1}/2'):
                pair.classify(masking=0)
                if pair.call == 'no_phase_change': # ignore if no phase change
                    continue
                else:
                    bed_writer.writerow({
                        '#chrom': pair.rec_1.reference_name,
                        'start': pair.condensed[0][1],
                        'end': pair.condensed[-1][2]}
                        )
        f.close()

        # reduce
        print('[readcomb] initial bed generated - reducing...')
        self.sort_tabix_bed(tabix_bed=False) # creates sorted bed without tabixing
        with open(self.bed_fname + 'reduced', 'w') as f_out:
            writer = csv.DictWriter(f_out, delimiter='\t', fieldnames=fieldnames)
            writer.writeheader()
            with open(self.bed_fname, 'r') as f_in:
                reader = csv.reader(f_in, delimiter='\t')
                _ = next(reader) # skip header
                prev_chrom, left_bound, prev_end = next(reader)
                left_bound, prev_end = int(left_bound), int(prev_end)

                # iterate through sorted lines
                for line in tqdm(reader, desc='reducing overlapping intervals'):
                    chrom = line[0]
                    start, end = [int(n) for n in line[1:]]
                    if not prev_chrom:
                        prev_chrom = chrom

                    # new chr started on line
                    if chrom != prev_chrom:
                        writer.writerow({
                            '#chrom': prev_chrom,
                            'start': left_bound,
                            'end': prev_end})
                        left_bound = start
                        prev_end = end
                        prev_chrom = None
                        continue
                    # current read overlaps with previous - extend
                    elif start <= prev_end <= end: 
                        prev_end = end
                    # current read does not overlap - write prev interval
                    elif prev_end < start: 
                        writer.writerow({
                            '#chrom': chrom,
                            'start': left_bound,
                            'end': prev_end})
                        left_bound = start
                        prev_end = end

        # fix up names
        os.remove(self.bed_fname)
        os.rename(self.bed_fname + 'reduced', self.bed_fname)


    def sort_tabix_bed(self, tabix_bed=True):
        """sort, uniq, and tabix bed file

        invokes subprocess for all three steps. the same method can be used on
        all three bed types

        tabix arg can be set to false if only sorting (used in nuclear bed
        generation)
        """
        # 'cat {in} | (sed -u 1q; sort -k1,1 -k2n,3n) | uniq > {out}'
        # write header to separate file
        header_file = open('header.temp', 'w')
        header_proc = subprocess.run(
            ['head', '-n', '1', self.bed_fname], stdout=header_file)
        header_file.close()

        # sort + uniq rest of file
        sorted_file = open('sorted.temp', 'w')
        header_remove_proc = subprocess.Popen(['tail', '-n', '+2', self.bed_fname], 
            stdout=subprocess.PIPE)
        sort_proc = subprocess.Popen(['sort', '-k1,1', '-k2n,3n'], 
            stdin=header_remove_proc.stdout,
            stdout=subprocess.PIPE)
        uniq_proc = subprocess.run(['uniq'], 
            stdin=sort_proc.stdout, stdout=sorted_file)
        sorted_file.close()

        # combine header and sorted file
        final_bed = open(self.bed_fname, 'w')
        cat_proc = subprocess.run(['cat', 'header.temp', 'sorted.temp'],
            stdout=final_bed)
        final_bed.close()
        print('[readcomb] created bed file')
        print('[readcomb] clearing temp files...')

        # clear temp files
        os.remove('header.temp')
        os.remove('sorted.temp')

        # bgzip and tabix file
        if tabix_bed:
            print('[readcomb] bgzip and tabix on bed file...')
            bgzip_proc = subprocess.run(['bgzip', self.bed_fname])
            tabix_proc = subprocess.run(['tabix', self.bed_fname + '.gz'])
            print('[readcomb] complete.')


class FalsePositiveFilterer():
    def __init__(self, fname, method, vcf_fname, tabix_reader, out, log):
        """Use generated BED file to filter out false positive reads

        Three available methods:

        - ``midpoint`` filtering will filter out putative recombinant reads that
          share at least one exact phase change with a false positive parental read
        - ``overlap`` filtering will compare variants common to a putative
          recombinant read and all overlapping false positive parental reads.
          even one instance of shared haplotypes will cause read filtering
        - ``nuclear`` filtering will remove all putative recombinant reads that
          have any overlap at all with a false positive read, regardless of the
          false positive's haplotype

        Parameters
        ----------
        fname : str
            path to putative recombinant reads from readcomb-filter
        method : str
            one of 'midpoint', 'overlap', or 'nuclear'
        vcf_fname : str
            path to VCF used to call recombination events
        tabix_reader : pysam.libctabix.TabixFile
            path to bed outfile
        out : str
            path to output SAM to write filtered reads to
        log : str
            path to log file
        """
        self.fname = fname
        self.method = method
        self.vcf_fname = vcf_fname
        self.tabix_reader = tabix_reader
        self.bed_fname = tabix_reader.filename.decode('utf-8')
        self.log = log

        # instantiate writer
        bam_template = pysam.AlignmentFile(self.fname, 'r')
        self.writer = pysam.AlignmentFile(out, 'wh', template=bam_template)

    def filter_false_positives(self):
        print(f'[readcomb] filtering {self.fname} with {self.method} method')
        print(f'[readcomb] using {self.bed_fname} as false positive lookup')

        # set up log
        needs_header = bool(not os.path.isfile(self.log))
        log_f = open(self.log, 'a', newline='')
        log_writer = csv.DictWriter(
            log_f, delimiter='\t', 
            fieldnames=[
                'time', 'total_pairs', 'pairs_kept', 
                'pairs_removed', 'no_phase_change', 
                'perc_kept', 'method', 'fname'])
        if needs_header:
            log_writer.writeheader()
        log_row = {
            'time': datetime.datetime.now(), 'total_pairs': 0, 'pairs_kept': 0, 
            'pairs_removed': 0, 'no_phase_change': 0, 'perc_kept': 0.0, 
            'method': self.method, 'fname': self.fname}

        # filter
        reader = pairs_creation(self.fname, self.vcf_fname)

        for pair in tqdm(reader, desc=f'filtering {self.fname}'):
            log_row['total_pairs'] += 1
            pair.classify(masking=0)
            if pair.call == 'no_phase_change':
                log_row['no_phase_change'] += 1
                continue
            chrom = pair.rec_1.reference_name
            start, end = pair.detection[0][1], pair.detection[-1][1]

            # handle tabix issues if organelle/contig region iterator can't be made
            try:
                false_lookup = list(set([
                    line for line in self.tabix_reader.fetch(chrom, start, end)]))
            except ValueError as e:
                if any(region in e.args[0] for region in ['cpDNA', 'mtDNA', 'contig']):
                    false_lookup = []
                else:
                    raise e

            if not false_lookup: # no false phase changes at all
                log_row['pairs_kept'] += 1 
                self.writer.write(pair.rec_1)
                self.writer.write(pair.rec_2)
                continue # no need to apply a filter at all

            # if false reads in vicinity - filter
            if self.method == 'midpoint':
                passed = self._filter_midpoint(pair, chrom, start, end, false_lookup)
            elif self.method == 'overlap':
                passed = self._filter_overlap(pair, chrom, start, end, false_lookup)
            elif self.method == 'nuclear':
                log_row['pairs_removed'] += 1
                continue # nuclear option means indiscriminately ignoring all overlapping reads
            else:
                raise ValueError(
                    f'method must be one of midpoint, overlap, or nuclear, not {self.method}')

            # if filter was applied and read did not match false read - write
            if passed:
                log_row['pairs_kept'] += 1
                self.writer.write(pair.rec_1)
                self.writer.write(pair.rec_2)
                continue
            elif not passed:
                log_row['pairs_removed'] += 1

        log_row['perc_kept'] = round(log_row['pairs_kept'] / log_row['total_pairs'], 3)
        log_writer.writerow(log_row)
                

    def _filter_midpoint(self, pair, chrom, start, end, false_lookup):
        # reduce pair.detection output for comparison
        detection_reduced = [(hap, pos) for hap, pos, base in pair.detection]
        # iterate over false reads
        for false_read in false_lookup:
            chrom, pos1, pos2, hap1, hap2, parent = false_read.split('\t')
            false_detection = [(hap1, int(pos1)), (hap2, int(pos2))]
            if all(var in detection_reduced for var in false_detection):
                return False # false phase change match found - exit loop
            else:
                continue
        return True # only returned if all false phase changes aren't matched


    def _filter_overlap(self, pair, chrom, start, end, false_lookup):
        for false_read in false_lookup:
            chrom, pos1, pos2, hap1, hap2, fp_detection = false_read.split('\t')
            false_detection = ast.literal_eval(fp_detection)
            overlap_sites = set(
                [v[1] for v in pair.detection]).intersection(
                    [v[1] for v in false_detection])
            detection_filtered = [
                var for var in pair.detection if var[1] in overlap_sites]
            if all(var in false_detection for var in pair.detection):
                return False # all overlapping sites are the same in the false read
            else:
                continue
        return True # only returned if all false reads in region aren't matched

            
def handle_bed(args):
    """Function that generates/handles BED file

    This function should be used to handle BED files instead of invoking
    BedGenerator directly.

    Parameters
    ----------
    args : argparse.ArgumentParser
        command line arguments

    Returns
    -------
    tabix_reader : pysam.libctabix.TabixFile
        TabixFile object for querying
    """
    if args.false_bed_in: # bed file already provided
        tabix_reader = pysam.TabixFile(args.false_bed_in)
        return tabix_reader

    # if not - generate lookup bed file
    if args.false_bed_out.endswith('.gz'):
        args.false_bed_out = args.false_bed_out.rstrip('.gz')
    bed_generator = BedGenerator(
        false_plus=args.false_plus, 
        false_minus=args.false_minus,
        vcf_fname=args.vcf,
        method=args.method,
        bed_fname=args.false_bed_out)

    bed_generator.generate_bed() # will use method specified above
    bed_generator.sort_tabix_bed()
    tabix_reader = pysam.TabixFile(args.false_bed_out + '.gz')
    return tabix_reader


def main():
    parser = arg_parser()
    args = parser.parse_args()
    tabix_reader = handle_bed(args)
    filterer = FalsePositiveFilterer(
        fname=args.fname,
        method=args.method,
        vcf_fname=args.vcf,
        tabix_reader=tabix_reader,
        out=args.out,
        log=args.log)
    filterer.filter_false_positives()
    print(f'[readcomb] written to {args.out}')


if __name__ == '__main__':
    main()

        




#!/usr/bin/env python3
"""
Tools for recombination event classification.

Unlike the other modules in readcomb, which should be run at the command line,
classification is intended to have its contents imported and used in a Python
environment.
"""

import itertools
import re
import numpy as np
import pysam
from cyvcf2 import VCF

try:
    from readcomb.filter import check_variants
    from readcomb.filter import cigar
    from readcomb.filter import qualities_cigar
except ImportError as e:
    print('WARNING: readcomb is not installed')
    print('Use command: pip install readcomb')
    from filter import check_variants
    from filter import cigar
    from filter import qualities_cigar

__version__ = '0.3.13'

def downstream_phase_detection(variants, segment, record, quality):
    """
    Detect haplotype of variants in the given DNA sequence.

    This function is exactly the same as ``phase_detection`` utilized in ``readcomb-filter``
    except it also returns the index of variants.

    ``downstream_phase_detection()`` takes a ``segment`` and a list of ``cyvcf2``
    variants in the region of the segment to detect the haplotype of the sequence and
    generate a list of tuples in the form ``[({'1', '2', 'N'}, variant_index), ...]`` to represent
    haplotypes in the order of the variants on the segment.

    Parameters
    ----------
    variants : list of cyvcf2.Variant
        ``cyvcf2`` variants in the bounds of ``segment`` given by ``check_variants()``
    segment : str
        realigned sequence built by ``cigar()``
    record : pysam.AlignedSegment
        bam record from pysam alignment file

    Returns
    -------
    list of tuples
        list of haplotypes of variants and their index in the bounds of the segment
    """

    detection_results = []

    cigar_tuples = record.cigartuples

    # check for no alignment
    if not cigar_tuples:
        return detection_results

    # get realigned base qualities
    query_qualities = qualities_cigar(record)

    for variant in variants:
        # Using SNP.start and record.reference_start since they are both 0 based
        # SNP.start grabs vcf positions in 0 index while vcfs are 1 indexed
        idx = variant.start - record.reference_start

        parent1 = variant.gt_bases[0].split('/')[0].split('|')[0]
        parent2 = variant.gt_bases[1].split('/')[0].split('|')[0]

        if parent1 == parent2: # ignore uninformative SNPs
            continue

        # ignore if snp is before sequence
        if idx < 0:
            continue

        # ignore variant if quality of sequencing at that base is below threshold
        if not query_qualities[idx]:
            continue # variant site is deleted in read

        if query_qualities[idx] < quality:
            continue

        if variant.is_indel:
            # check if indel is outside of segment
            if idx + max(len(parent1), len(parent2)) > len(segment):
                continue

            parent1_match = segment[idx:idx + len(parent1)] == parent1
            parent2_match = segment[idx:idx + len(parent2)] == parent2

            if parent1_match and not parent2_match:
                detection_results.append(('1', variant.start, parent1))
            elif parent2_match and not parent1_match:
                detection_results.append(('2', variant.start, parent2))
            elif parent1_match and parent2_match:
                continue
            else:
                detection_results.append(('N', variant.start, None))

        else: # variant is a SNP
            if idx >= len(segment):
                break

            if segment[idx] == parent1:
                detection_results.append(('1', variant.start, parent1))

            elif segment[idx] == parent2:
                detection_results.append(('2', variant.start, parent2))

            else:
                detection_results.append(('N', variant.start, None))

    return detection_results

class Pair():
    def __init__(self, record1, record2, vcf_filepath):
        """
        Initialize a ``Pair`` object

        It is highly recommended to use ``classification.pairs_creation`` to create pairs by
        parsing through a SAM/BAM file.

        ``record1`` and ``record2`` are reordered so that they are in increasing sequential order

        Parameters
        ----------
        record1: pysam.AlignedSegment
            one of the bam sequences in a read pair
        record2: pysam.AlignedSegment
            bam sequence in a read pair corresponding to the one in ``record1``
        vcf_filepath: str
            filepath to VCF file that contains variants for ``record1`` and ``record2``
        """
        # put pairs in order by reference_start of the bam sequences
        if record1.reference_start < record2.reference_start:
            self.rec_1 = record1
            self.rec_2 = record2
        elif (
            record1.reference_start == record2.reference_start and
            record1.reference_end > record2.reference_end # rare, but possible
        ):
            self.rec_1 = record2
            self.rec_2 = record1
        else:
            self.rec_1 = record2
            self.rec_2 = record1

        self.vcf_filepath = vcf_filepath

        # descriptive attributes for pairs
        self.segment_1 = cigar(self.rec_1)
        self.segment_2 = cigar(self.rec_2)
        self.variants_1 = None
        self.variants_2 = None
        self.variants_all = None
        self.variants_filt = None
        self.midpoint = None
        self.relative_midpoint = -1
        self.location = ''.join(f'{self.rec_1.reference_name}:\
        {self.rec_1.reference_start}-\
        {self.rec_2.reference_start + len(self.segment_2)}'.split())

        # recombination event calling
        self.detection_1 = None
        self.detection_2 = None
        self.detection = None
        self.masked_detection = None
        self.no_match = None
        self.condensed = None
        self.masked_condensed = None
        self.call = None
        self.masked_call = None
        self.gene_conversion_len = None

        # quality metrics
        self.overlap = None
        self.overlap_disagree = None
        self.variants_per_haplotype = -1
        self.min_variants_in_haplotype = -1
        self.variant_skew = -1
        self.mismatch_variant_ratio = -1
        self.variant_counts = None
        self.converted_haplotype = None
        self.converted_variants = None
        self.indel_proximity = None

    def __str__(self):
        """
        Generate string representation of ``Pair`` class.

        This function overwites the built-in ``__str__`` function of ``Pair``.
        The ``__str__`` function is called whenever a user prints a ``Pair`` object or
        converts the object to a ``str``. ``Pair`` has two different states, before and
        after calling ``classify()``, so two different string representations are
        needed.

        Returns
        -------
        string : str
            string representation of the ``Pair`` object
        """
        if self.call:
            string = f'Record name: {self.rec_1.query_name} \n' + \
                    f'Read1: {self.rec_1.reference_name}:{self.rec_1.reference_start}' + \
                    f'-{self.rec_1.reference_start + len(self.segment_1)} \n' + \
                    f'Read2: {self.rec_2.reference_name}:{self.rec_2.reference_start}' + \
                    f'-{self.rec_2.reference_start + len(self.segment_2)} \n' + \
                    f'VCF: {self.vcf_filepath} \n' + \
                    f'Unmatched Variant(s): {self.no_match} \n' + \
                    f'Condensed: {self.condensed} \n' + \
                    f'Call: {self.call} \n' + \
                    f'Condensed Masked: {self.masked_condensed} \n' + \
                    f'Masked Call: {self.masked_call} \n' + \
                    f'Midpoint: {self.get_midpoint()} \n' + \
                    f'Variants Per Haplotype: {self.variants_per_haplotype} \n' + \
                    f'Variant Skew: {self.variant_skew} \n' + \
                    f'Gene Conversion Length: {self.gene_conversion_len} \n' + \
                    f'Min Variants In Haplotype: {self.min_variants_in_haplotype} \n' + \
                    f'Mismatch/Variant Ratio: {self.mismatch_variant_ratio} \n'
        else:
            string = f'Record name: {self.rec_1.query_name} \n' + \
                    f'Read1: {self.rec_1.reference_name}:{self.rec_1.reference_start}' + \
                    f'-{self.rec_1.reference_start + len(self.segment_1)} \n' + \
                    f'Read2: {self.rec_2.reference_name}:{self.rec_2.reference_start}' + \
                    f'-{self.rec_2.reference_start + len(self.segment_2)} \n' + \
                    f'VCF: {self.vcf_filepath}'

        return string

    def package(self):
        """
        Convert ``self.rec_1`` and ``self.rec_2`` from ``pysam.AlignedSegment`` to ``str``

        The user may want to implement multiprocessing to decrease the amount
        of time to classify all reads in a SAM/BAM file. ``self.rec_1`` and
        self.``rec_2`` and all ``pysam.AlignedSegment`` objects are not pickleable and
        cannot be passed through a ``multiprocessing.Queue``. Instead of directly
        handling BAM/SAM strings, users can choose to create a ``Pair``, call
        ``package`` to convert the records to strings using the ``to_string()``
        function from ``pysam`` and pass the ``Pair`` object through a ``Queue``.
        """
        if (
            type(self.rec_1) == type(pysam.AlignedSegment()) and
            type(self.rec_2) == type(pysam.AlignedSegment())
        ):
            self.rec_1 = self.rec_1.to_string()
            self.rec_2 = self.rec_2.to_string()

    def unpackage(self, header):
        """
        Convert ``self.rec_1`` and ``self.rec_2`` from ``str`` to ``pysam.AlignedSegment``

        If the user is implementing multiprocessing to decrease the amount of
        time to classify all reads in a SAM/BAM file, the ``package()`` function should
        be called on ``Pair`` to convert ``self.rec_1`` and ``self.rec_2`` to ``str``
        and passed to a subprocess. After the subprocess receives the packaged
        ``Pair``, it should be unpackaged using this function.

        ``pysam.AlignedSegment.fromstring()`` also requires the ``pysam.AlignmentHeader``
        object which can be obtained from calling ``.header`` on the
        ``pysam.AlignmentFile`` object that this ``Pair`` was parsed from using
        ``pair_creation()``.

        Parameters
        ----------
        header : pysam.AlignmentHeader
            ``pysam.AlignmentHeader`` of the ``pysam.AlignmentFile`` associated with this ``Pair``
        """
        if isinstance(self.rec_1, str) and isinstance(self.rec_1, str):
            self.rec_1 = pysam.AlignedSegment.fromstring(self.rec_1, header)
            self.rec_2 = pysam.AlignedSegment.fromstring(self.rec_2, header)

    def classify(self, masking=0, quality=30, vcf=None):
        """
        Determine the type of recombination event that occursed in this read pair.

        ``classify()`` does phase change analysis similar to ``readcomb-filter`` with addtional
        steps to classify the type of recombination event. See ``readcomb.check_variants()``,
        ``readcomb.cigar()``, and ``downstream_phase_detection`` for more information on the
        analysis process.

        Classification is done on the haplotypes present in the full read pair and also when the
        beginning and end are shortened by ``masking`` bases as phase changes close to the start
        and end of a read pair are difficult to call as crossovers or gene conversions. The two
        different classifications allows a more nuanced call on the phase change type present in
        the pair.

        ``classify()`` also takes in optional ``vcf`` for a ``cyvcf2.VCF`` object. It is highly
        recommended to pass in a pre-initialized ``VCF`` object when large amounts of ``Pair``
        objects are being classified as the creation of ``VCF`` parsers greatly affects
        performance and tends to be less reliable when large amounts of parsers are created.

        Do not use the same ``cyvcf2.VCF`` parser across multiple processes/threads as it leads to
        errors involving file access permissions. Use a dedicated ``cyvcf2.VCF`` parser for
        each process/thread instead.

        Parameters
        ----------
        masking : int, default 0
            number of bases to be ignored when determing the ``masked_classification``
        quality : int, default 30
            filter quality for individual bases in a sequence
        vcf: cyvcf2.VCF, optional
            pre-initialized ``cyvcf2.VCF`` for this ``Pair`` to use
            when classifying, highly recommended
        """
        # create new VCF object if none is provided
        if not vcf:
            vcf = VCF(self.vcf_filepath)

        self.variants_1 = check_variants(
            vcf, self.rec_1.reference_name, self.rec_1.reference_start, 
            self.rec_1.reference_start + len(self.segment_1))
        self.variants_2 = check_variants(
            vcf, self.rec_2.reference_name, self.rec_2.reference_start, 
            self.rec_2.reference_start + len(self.segment_2))
        vcf.close()

        self.detection_1 = downstream_phase_detection(
            self.variants_1, self.segment_1, self.rec_1, quality)
        self.detection_2 = downstream_phase_detection(
            self.variants_2, self.segment_2, self.rec_2, quality)

        # deal with 'het pairs' and 'duplicate variants' - see _resolve_overlap() documentation
        # updates the read-specific detection methods
        self._resolve_overlap()
        self.detection = sorted(self.detection_1 + self.detection_2,
            key=lambda variant: variant[1]) # sort by position

        # create condensed variant list for easy look up
        # this hideous listcomp removes duplicates and sorts variants
        self.variants_all = [
            [variant for variant in self.variants_1 + self.variants_2 if variant.POS == pos][0]
            for pos in sorted(set(variant.POS for variant in self.variants_1 + self.variants_2))]

        # create filtered variant list only containing variants used for detection
        self.variants_filt = [
            variant for variant in self.variants_all
            if variant.POS in [v[1] + 1 for v in self.detection]]
            
        # set no_match variable if there are unmatched variants
        self.no_match = any([v[0] == 'N' for v in self.detection])

        # simplification of results - creating condensed hap view
        # [(haplotype, beginning, end), ...]
        self.condensed = self._create_condensed(self.detection)

        # create list of just haplotype information without range from condensed
        haplotypes = [tupl[0] for tupl in self.condensed if tupl[0] != 'N']

        # classify condensed
        if len(haplotypes) == 2:
            self.call = 'cross_over'
        elif len(haplotypes) == 3:
            self.call = 'gene_conversion'
        elif len(haplotypes) > 3:
            self.call = 'complex'
        else:
            self.call = 'no_phase_change'

        if self.rec_2.reference_start + len(self.segment_2) \
            - self.rec_1.reference_start < masking * 2:

            print('Masking size too large for pair: ' + self.rec_1.query_name)
            self.masked_condensed = None
            self.masked_call = None
            return

        mask_start = self.rec_1.reference_start + masking
        mask_end = self.rec_2.reference_start + len(self.segment_2) - masking
        self.masked_detection = [
            v for v in self.detection if mask_start < v[1] < mask_end]

        # create masked_condensed from masked detection
        # [(haplotype, beginning, end), ...]
        self.masked_condensed = self._create_condensed(self.masked_detection)

        # create list of just haplotype information no range from condensed
        masked_haplotypes = [tupl[0] for tupl in self.masked_condensed if tupl[0] != 'N']

        # classify masked_condensed
        if len(masked_haplotypes) == 2:
            self.masked_call = 'cross_over'
        elif len(masked_haplotypes) == 3:
            self.masked_call = 'gene_conversion'
        elif len(masked_haplotypes) > 3:
            self.masked_call = 'complex'
        else:
            self.masked_call = 'no_phase_change'

        # convert haplotypes 1/2/N to VCF sample names
        samples = vcf.samples

        for tupl in self.condensed:
            if tupl[0] == '1':
                tupl[0] = samples[0]
            elif tupl[0] == '2':
                tupl[0] = samples[1]

        for tupl in self.masked_condensed:
            if tupl[0] == '1':
                tupl[0] = samples[0]
            elif tupl[0] == '2':
                tupl[0] = samples[1]

        # set descriptive + quality attributes post-classification
        self._describe(haplotypes, vcf)

        # if gene conversion - set descriptive info
        if 'gene_conversion' in [self.call, self.masked_call]:
            self._describe_gene_conversion(vcf)
        else:
            self.gene_conversion_len = 'N/A'
            self.converted_haplotype = 'N/A'
            self.converted_variants = 'N/A'

        self.midpoint, self.relative_midpoint = self.get_midpoint()

    def _create_condensed(self, detection):
        """
        Internal method to generate haplotype representation in
        ``Pair.condensed``.

        Parameters
        ----------
        detection : list
            list containing variants in downstream_phase_detection format
        """
        condensed = []

        for variant in detection:
            haplotype, location, base = variant

            # first variant
            if len(condensed) == 0:
                condensed.append(
                    [haplotype, self.rec_1.reference_start, self.rec_1.reference_start])

            # if same haplotype - extend segment for correct eventual midpoint calc
            elif condensed[-1][0] == haplotype:
                condensed[-1][2] = location

            # different haplotype
            elif condensed[-1][0] != haplotype:
                # middle of previous variant location and current variant
                # rounding down so it's not a decimal
                midpoint = int((condensed[-1][2] + location) // 2)
                condensed[-1][2] = midpoint
                condensed.append([haplotype, midpoint, midpoint])

            # last variant
            if len(self.detection_2) == 0:
                if variant == self.detection_1[-1]:
                    condensed[-1][2] = self.rec_1.reference_start + \
                        len(self.segment_1)
            else:
                if variant == self.detection_2[-1]:
                    condensed[-1][2] = self.rec_2.reference_start + \
                        len(self.segment_2)

        # this occurs during false positive phase changes when reads overlap
        if any(start == end for hap, start, end in condensed):
            for false_hap in [h for h in condensed if h[1] == h[2]]:
                condensed.remove(false_hap)
        
        return condensed

    def _describe(self, haplotypes, vcf):
        """
        Internal method to calculate various attributes of ``Pair``. Sets
        the following attributes:

        - mismatch_variant_ratio - number of mismatches to reference over the
          number of variants - see ``Pair._get_mismatch_variant_ratio()``
          documentation for details
        - variants_per_haplotype - average number of variants supporting each
          haplotype
        - min_variants_in_haplotype - number of variants in least
          well-supported haplotype
        - outer_bound - relative location of 'most outer' variant involved
          in a phase change

        Parameters
        ----------
        haplotypes : list
            haplotype listing generated from self.masked_condensed
        vcf : cyvcf2.VCF
            VCF file object generated by ``classify()``
        """
        if not self.call:
            self.classify()

        # use helper function to calculate unexpected mismatch count
        self.mismatch_variant_ratio = self._get_mismatch_variant_ratio()

        # calculate average number of variants per haplotype and create dict of counts
        self.variants_per_haplotype = len(self.variants_filt) / max(len(haplotypes), 1)
        self.variant_counts = {
            hap: [t[0] for t in self.detection].count(hap) for hap in set(haplotypes)}
        if len(self.variant_counts.values()) == 2: # only calc if both haps present
            self.variant_skew = max(self.variant_counts.values()) / min(self.variant_counts.values())

        # get the lowest number of variants a haplotype has -
        # splits variant list (e.g. ['1', '1', '2', '1']) and gets min variant count across haps
        if 'no_phase_change' not in set([self.call, self.masked_call]):
            self.min_variants_in_haplotype = min(
                len(list(grouper)) for value, grouper
                in itertools.groupby([hap for hap, position, allele in self.detection])
                )

        if self.call != 'no_phase_change':
            phase_change_variants = [
                [self.detection[i][1], self.detection[i+1][1]]
                for i in range(len(self.detection) - 1)
                if self.detection[i][0] != self.detection[i+1][0]]
            phase_change_variants = sorted(list(set(
                [pos for var_pair in phase_change_variants for pos in var_pair]
            )))
            # only check leftmost and rightmost variants
            if (
                phase_change_variants[0] - self.condensed[0][1] < \
                abs(phase_change_variants[-1] - self.condensed[-1][2])
            ):
                self.outer_bound = round(
                    (phase_change_variants[0] - self.rec_1.reference_start) / \
                    (self.condensed[-1][2] - self.condensed[0][1]), 3)
            else:
                self.outer_bound = round(
                    (phase_change_variants[-1] - self.rec_1.reference_start) / \
                    (self.condensed[-1][2] - self.condensed[0][1]), 3)
            

    def _describe_gene_conversion(self, vcf):
        """
        Internal method to calculate various attributes of ``Pair`` if the
        pair is called as a gene conversion. Sets the following attributes:

        The following attributes are only set if the read pair is called as a
        gene conversion:
        - gene_conversion_len - length of gene conversion segment, if applicable,
          as determined by midpoint method
        - converted_haplotype - which haplotype was 'converted to'
        - converted_variants - which variants were converted in the tract
        - indel_proximity - how close the nearest indel is to the converted tract, 
          if any
        """
        self.gene_conversion_len = self.condensed[-1][1] - self.condensed[0][2]

        # converted haplotype and variants
        self.converted_haplotype = [
            hap for hap, count in self.variant_counts.items() 
            if count == min(self.variant_counts.values())][0]
        self.converted_variants = [
            variant for variant in self.detection if variant[0] == self.converted_haplotype]

        # set converted haplotype to sample name
        if self.converted_haplotype == '1':
            self.converted_haplotype = vcf.samples[0]
        elif self.converted_haplotype == '2':
            self.converted_haplotype = vcf.samples[1]

        # indel proximity
        indels = []
        for rec in [self.rec_1, self.rec_2]:
            idx = rec.reference_start
            for op, basecount in rec.cigartuples:
                if op == 0:
                    idx += basecount
                elif op == 1:
                    indels.append(['ins', idx, idx])
                elif op == 2:
                    indels.append(['del', idx, idx + basecount])
        # overlapping reads may have the same indel register twice
        indels = list(indel for indel, _ in itertools.groupby(indels)) 

        conversion_tract = np.array([
            tract for tract in self.condensed 
            if tract[0] == self.converted_haplotype][0][1:])
        if indels:
            self.indel_proximity = min(abs(np.concatenate(
                [np.array(indel[1:]) - np.array(conversion_tract)
                 for indel in indels])))
        else:
            self.indel_proximity = 'N/A'
        

    def _get_mismatch_variant_ratio(self):
        """
        Helper function to calculate mismatch_variant_ratio quality metric. Will
        use MD tags in reads to get the number of mismatches to the reference
        per read, and then compare with the number of variants. Ratios above 1
        indicate variation that was either deemed too low quality to be
        included in the VCF. It is possible some sites could have diverged from
        the reference and resulted in the same allele in both parents, but this
        is less likely than the region simply being prone to paralogous
        alignment if the ratio is particularly high.

        The MD tag also includes deletions, designated as '^ATG' if 'ATG' is a
        3 bp deletion from the reference in the read. These deletions are
        removed from consideration.
        """
        if len(self.variants_1) + len(self.variants_2) == 0:
            return -1
        if len(self.variants_filt) == 0:
            return -1

        if not self.overlap:
            rec_1_mismatches = len(
                re.findall(
                    r'[ACGT]', re.sub(r'\^[ACGT]+', '', self.rec_1.get_tag('MD'))
                    )
                )
            rec_2_mismatches = len(
                re.findall(
                    r'[ACGT]', re.sub(r'\^[ACGT]+', '', self.rec_2.get_tag('MD'))
                    )
                )
            return (rec_1_mismatches + rec_2_mismatches) / len(self.variants_filt)

        elif self.overlap:
            # if rec2 is entirely 'contained' by rec1 - just do rec1
            if (
                self.rec_1.reference_start <= self.rec_2.reference_start and
                self.rec_2.reference_end < self.rec_1.reference_end
            ):
                m1 = self.rec_1.get_tag('MD')
                mismatches = len(
                    re.findall(
                        r'[ACGT]', re.sub(r'\^[ACGT]+', '', self.rec_1.get_tag('MD'))
                        )
                    )
                return mismatches / len(self.variants_filt)

            # otherwise - handle overlap as normal
            m1 = self.rec_1.get_tag('MD')
            m2 = self.rec_2.get_tag('MD')
            left_region = None
            right_region = None

            # get overlap
            for i, _ in enumerate(m1):
                # remove up to 3 digits from end
                # since these may differ
                subseq = re.sub(r'[0-9]{0,3}$', '', m1[i:])
                # test subseqs
                if subseq in m2 and len(subseq) != 1: # not last possible base
                    overlap_seq = subseq
                    break
            if not overlap_seq: # happens if the segment removed in re.sub above was the overlap
                if m2 in m1[-3:]:
                    overlap_seq = m2
                else:
                    # overlap still not detected - just keep both
                    left_region = m1
                    right_region = m2

            # get regions on either side of overlap if not assigned just above
            if not left_region and not right_region:
                left_region = m1[:m1.index(overlap_seq)]
                right_region = re.sub(
                    r'^[0-9]{0,3}', '',
                    m2[m2.index(overlap_seq) + len(overlap_seq):]
                )
            mismatches = 0
            for region in [left_region, overlap_seq, right_region]:
                mismatches += len(
                    re.findall(
                        r'[ACGT]', re.sub(r'\^[ACGT]+', '', region)
                        )
                    )
            return mismatches / len(self.variants_filt)

    def _resolve_overlap(self):
        """
        Helper function for ``Pair.classify()``. 

        Resolves cases where the same site is sequenced in both reads (i.e. the
        reads overlap) but the calls at the two sites differ. This will
        otherwise be incorrectly flagged as a phase change when it's more
        likely to be a sequencing error or paralogous alignment. This method
        filters the variants_* and detection_* attributes to remove any
        variants where this is the case.

        If the variants do have the same call, the variant in the second (3')
        read will be removed so that the call doesn't show multiple times in
        ``Pair.detection``.

        Will set the self.overlap attribute to True if variants overlap.
        Finally, will check whether there are discrepancies in base calls
        between overlapping reads at all (and will set overlap_disagree
        accordingly).
        """
        # check variant site overlap first
        overlapping_sites = list(set(var[1] for var in self.detection_1).intersection(
            set(var[1] for var in self.detection_2)))

        # set self.overlap to True if variant sites overlap
        if overlapping_sites and not self.overlap:
            self.overlap = True

        for site in overlapping_sites:
            read_1_call = [(hap, pos, base) for hap, pos, base in self.detection_1 if pos == site][0]
            read_2_call = [(hap, pos, base) for hap, pos, base in self.detection_2 if pos == site][0]

            # if calls at same site disagree, remove variant from consideration
            if read_1_call != read_2_call:
                self.detection_1.remove(read_1_call)
                self.detection_2.remove(read_2_call)
            elif read_1_call == read_2_call:
                self.detection_2.remove(read_2_call) # only remove from right read

        # check if there is any overlap at all
        if not self.overlap:
            if self.rec_2.reference_start < self.rec_1.reference_start + len(self.segment_1):
                self.overlap = True
            else:
                self.overlap = False

        # if there's overlap, check whether overlapping sites anywhere in read disagree
        if self.overlap:
            # identify overlapping region
            region = [self.rec_2.reference_start, self.rec_1.reference_start + len(self.segment_1)]
            idx = self.rec_2.reference_start - self.rec_1.reference_start
            segment_1 = self.segment_1[idx:region[1]]
            segment_2 = self.segment_2[0:len(segment_1)]
            if segment_1 != segment_2: # compare full strings
                self.overlap_disagree = False
            else:
                self.overlap_disagree = True

    def get_midpoint(self):
        """
        Determine and return the midpoint of ``Pair``

        The midpoint of a read pair with no phase changes is halfway between the start
        of the first read and end of second read.

        The midpoint of a read pair with a phase change is halfway between the two closest
        variants that signify a phase change event. For gene conversions, it is halfway between
        the two outer variants of the group of 3. This logic extends to read pairs
        with complex haplotypes.

        Returns
        -------
        midpoint : int
            the middle point of the phase change event of the ``Pair``
        relative_midpoint : float
            the location of the midpoint relative to the length of the ``Pair``
        """
        # give error if Pair object is packaged
        if isinstance(self.rec_1, str) or isinstance(self.rec_2, str):
            raise TypeError('cannot classify if Pair object is not unpackaged yet')

        # return midpoint if it's already been called
        if self.midpoint:
            return self.midpoint

        # classify if read has not been already
        if not hasattr(self, 'call'):
            self.classify()

        # simplification of results
        # [(haplotype, beginning, end), ...]

        if self.masked_call == 'no_phase_change':
            # midpoint is middle of two paired reads if no phase change
            start = self.rec_1.reference_start
            end = self.rec_2.reference_start + len(self.segment_2)
            self.midpoint = int((start + end) / 2)
            self.relative_midpoint = -1
        elif self.masked_call == 'cross_over':
            # (X, begin, end), (Y, begin, end): end of X = beginning of Y = midpoint
            self.midpoint = self.condensed[0][2]
            # get relative midpoint
            start = self.condensed[0][1]
            end = self.condensed[-1][2]
            self.relative_midpoint = round(
                (self.midpoint - self.rec_1.reference_start) / (end - start), 3)
        elif self.masked_call == 'gene_conversion':
            # (X, begin, end), (Y, begin, end), (X', begin, end)
            # midpoint of Y tract
            self.midpoint = sum([
                [start, end] for hap, start, end in self.condensed
                if hap == self.converted_haplotype][0]) / 2
            start = self.condensed[0][1]
            end = self.condensed[-1][2]
            self.relative_midpoint = round(
                (self.midpoint - self.rec_1.reference_start) / (end - start), 3)
        else:
            # kind of a shortcut using midpoints to find the middle
            start = self.condensed[0][2]
            end = self.condensed[-1][1]
            self.midpoint = (start + end) / 2
            self.relative_midpoint = -1

        # return in samtools format
        self.midpoint = f'{self.rec_1.reference_name}:{round(self.midpoint)}'

        return self.midpoint, self.relative_midpoint


def pairs_creation(bam_filepath, vcf_filepath):
    """
    Parses through a BAM/SAM file and generates ``Pair`` objects for read pairs.

    Given a read name sorted BAM/SAM file, ``pairs_creation`` creates ``Pair``
    objects with read pairs.

    It is recommended to use this function to create ``Pair`` objects for an entire
    BAM/SAM file instead of manually creating the ``Pair``. It is also recommended
    to use ``pairs_creation`` on SAM files that have already been filtered by
    ``readcomb-filter`` instead of full BAMs/SAMs.

    Parameters
    ----------
    bam_filepath : str
        filepath of BAM/SAM file processed with ``readcomb-bamprep``
    vcf_filepath : str
        filepath of VCF files that contain variants for reads in ``bam_filepath``

    Returns
    -------
    pairs : generator
        generator that yields reads from ``bam_filepath`` in the form of ``Pair`` objects
    """
    bam = pysam.AlignmentFile(bam_filepath, 'r')

    prev_rec = None
    for rec in bam:
        # first record
        if not prev_rec:
            prev_rec = rec
        # check if query_name pairs exist
        elif rec.query_name == prev_rec.query_name:
            yield Pair(prev_rec, rec, vcf_filepath)
        elif rec.query_name != prev_rec.query_name:
            prev_rec = rec
            continue



#!/usr/bin/env python3
"""
Readcomb filtering script utilizing multiprocessing
"""
import os
import sys
import argparse
import time
import datetime
import subprocess
from io import BytesIO
from multiprocessing import Queue
from multiprocessing import Process

import pysam
import pandas as pd
from cyvcf2 import VCF
from tqdm import tqdm

class ReadcombParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f'error: {message}\n')
        self.print_help()
        sys.exit(1)


def arg_parser():
    parser = ReadcombParser(
        description='filter BAM for reads containing phase changes',
        usage='readcomb-filter [options]')

    parser.add_argument('-b', '--bam', required=True,
        type=str, help='BAM to filter, required')
    parser.add_argument('-v', '--vcf', required=True,
        type=str, help='VCF containing parents, required')
    parser.add_argument('-p', '--processes', required=False, type=int, default=4,
        help='Number of processes to run readcomb filter on, default is 1')
    parser.add_argument('-m', '--mode', required=False, type=str, default='phase_change',
        help='Read filtering mode (default phase_change) [phase_change|no_match]')
    parser.add_argument('-l', '--log', required=False, default=None,
        type=str, help='Filename for log metric output [optional]')
    parser.add_argument('-o', '--out', required=False, type=str, default='recomb_diagnosis',
        help='File to write to (default recomb_diagnosis)')
    parser.add_argument('-q', '--quality', required=False, type=int, default=30,
        help='Filter quality for individual bases in a sequence (default: 30)')
    parser.add_argument('--min_mapq', required=False, type=int, default=40,
        help='Minimum MAPQ for read consideration (default: 40)')
    parser.add_argument('--version', action='version', version='readcomb 0.3.13')

    return parser

class SilentVCF:
    # helper class for check_variants that silences cyvcf2 warnings
    def __enter__(self):
        self._original_stderr = sys.stderr
        sys.stderr = open(os.devnull, 'a')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stderr.close()
        sys.stderr = self._original_stderr

def check_variants(vcf_file_obj, chromosome, left_bound, right_bound):
    """
    Fetch variants from VCF file in given bounds.

    Parameters
    ----------
    vcf_file_obj : cyvcf2.VCF
        ``cyvcf2`` VCF file object
    chromosome : str
        reference sequence name of bound
    left_bound : int
        0-based index of sequence left bound
    right_bound : int
        0-based index of sequence right bound

    Returns
    -------
    list of cyvcf.Variant
    """

    # 1 is added to record.reference_start and the following parameter because vcf is 1 indexed
    # in order to keep code consistent
    region = f'{chromosome}:{left_bound+1}-{right_bound}'
    with SilentVCF():
        records = [record for record in vcf_file_obj(region)]
    return records

def cigar(record):
    """
    Realign the given query segment to its reference.

    Rebuild the sequence using the query sequence and cigar tuple from the
    bam record to realign it with the reference sequence prior to variant matching

    Parameters
    ----------
    record : pysam.AlignedSegment
        bam record from pysam alignment file

    Returns
    -------
    str
        a sequence of bases representing the realigned DNA sequence
    """
    cigar_tuples = record.cigartuples

    # initialize segment for building
    segment = []

    # index to keep track of where we are in the query_segment
    query_segment = record.query_sequence
    index = 0

    # no alignment, cigar_tuple is equal to None
    if cigar_tuples is None:
        return segment

    # terminology on consume query/reference from samtools SAM specifications
    for cigar_tuple in cigar_tuples:

        # 0 is a match, add it onto the segment
        # consume query and reference
        if cigar_tuple[0] == 0:
            segment.append(query_segment[index:index+cigar_tuple[1]])
            index += cigar_tuple[1]

        # 1 is an insertion in query segment
        # consume query by increasing index
        elif cigar_tuple[0] == 1:
            index += cigar_tuple[1]

        # 2 is an deletion to query segment
        # consume reference by adding gaps to segment
        elif cigar_tuple[0] == 2:
            segment.append('-' * cigar_tuple[1])

        # 4 is soft clipping
        # consume query by increasing index
        elif cigar_tuple[0] == 4:
            index += cigar_tuple[1]

        # 5 = hard clipping
        # does NOT consume query or reference
        elif cigar_tuple[0] == 5:
            continue

        else:
            # currently does not support padding (6),
            # sequence match (7), and sequence mismatch (8)
            raise ValueError(
                'No condition for tuple {} of {}'.format(
                    cigar_tuples.index(cigar_tuple), cigar_tuples))

    return ''.join(segment)

def qualities_cigar(record):
    """
    Generate updated list of base qualities for use in ``phase_detection``.
    'Realigns' query qualities based on indels as needed. Follows the logic found
    in the ``cigar`` function.

    Deletions in the read relative to the reference are denoted as NoneType
    whereas insertions to the read are removed from consideration.

    For example, if a read has the cigar string is 10M10I10M10D10M (10 matches,
    10 insertions to read, 10 matches, 10 deletions from read, 10 matches, the
    realigned segment would include the matches, remove the inserted bases, include
    subsequent matches, add 10 '-' characters to reflect the deletion, and then
    include the final 10 bases.

    Similarly, this function will remove base qualities for inserted bases and
    insert None for each '-' character, so that indices are preserved when
    using ``phase_detection``.

    Parameters
    ----------
    record : pysam.AlignedSegment
        bam record from pysam alignment file

    Returns
    -------
    list of base qualities
        list of integers encoding base qualities
    """
    input_quals = record.query_qualities.tolist()

    # if all bases match, return full query qualities
    if all([op == 0 for op, basecount in record.cigartuples]):
        return input_quals

    idx = 0
    output_quals = []
    for op, basecount in record.cigartuples:

        # match - append qualities as normal
        if op == 0:
            output_quals.extend(input_quals[idx:idx + basecount])
            idx += basecount

        # insertion to the read - increment index and ignore
        elif op == 1:
            idx += basecount

        # deletion in the read - add None to quals for consistency with segment
        elif op == 2:
            output_quals.extend([None] * basecount)

        # soft clipping - increment index and ignore
        elif op == 4:
            idx += basecount

    return output_quals

def phase_detection(variants, segment, record, args):
    """
    Detect haplotype of variants in the given DNA sequence.

    Takes a ``segment`` and a list of ``cyvcf2`` variants in the region of the segment
    to detect the haplotype of the sequence and generate a list of {'1','2','N'} to
    represent it in the order of the variants on the segment

    Parameters
    ----------
    variants : list of cyvcf2.Variant
        ``cyvcf2`` variants in the bounds of ``segment`` given by ``check_variants()``
    segment : str
        realigned sequence built by ``cigar()``
    record : pysam.AlignedSegment
        bam record from pysam alignment file
    args : Namespace
        Namespace containing all user given arguments compiled by arg_parse()

    Returns
    -------
    list of {'1','2','N'}
        list of haplotypes of variants in the bounds of the segment
    """

    variant_lst = []

    cigar_tuples = record.cigartuples

    # check for no alignment
    if not cigar_tuples:
        return variant_lst

    query_qualities = qualities_cigar(record)

    for variant in variants:
        # Using variant.start and record.reference_start since they are both 0 based
        # variant.start grabs vcf positions in 0 index while vcfs are 1 indexed
        idx = variant.start - record.reference_start

        parent1 = variant.gt_bases[0].split('/')[0].split('|')[0]
        parent2 = variant.gt_bases[1].split('/')[0].split('|')[0]

        # ignore uninformative variants
        if parent1 == parent2:
            continue

        # ignore if variant is before sequence
        if idx < 0:
            continue

        # check if variant is outside of segment
        if idx + max(len(parent1), len(parent2)) > len(segment):
            continue

        # ignore variant if quality of sequencing at that base is below threshold
        if not query_qualities[idx]:
            continue # variant site is deleted in read

        if query_qualities[idx] < args.quality:
            continue

        if variant.is_indel:
            parent1_match = segment[idx:idx + len(parent1)] == parent1
            parent2_match = segment[idx:idx + len(parent2)] == parent2

            if parent1_match and not parent2_match:
                variant_lst.append('1')
            elif parent2_match and not parent1_match:
                variant_lst.append('2')
            elif parent1_match and parent2_match:
                continue # uninformative allele! 
            else:
                variant_lst.append('N')

        else: # variant is a SNP
            if segment[idx] == parent1:
                variant_lst.append('1')

            elif segment[idx] == parent2:
                variant_lst.append('2')

            else:
                variant_lst.append('N')

    return variant_lst

class Processor(Process):
    def __init__(self, input_queue, counter_queue,
                writer_queue, args, **kwargs):
        """
        Initialize the read pair processor.

        Processor is inherited from ``multiprocessing.Process``; when started it receives
        pairs of bam sequences through the ``input_queue`` from the scheduler and does
        phase change analysis, outputting through ``writer_queue`` bams that fit the user
        given argument criteria to ``Writer``.

        When the scheduler has parsed and distributed all bam sequences, the ``Processor``
        receives a ``None`` through the ``input_queue`` and ends its loop and prepares to join.

        Parameters
        ----------
        input_queue : multiprocessing.Queue
            input queue of bam read pairs in string form from scheduler
        counter_queue : multiprocessing.Queue
            output queue for result of phase change analysis
        writer_queue : multiprocessing.Queue
            output queue for bam read pairs in string form that fulfill the user given criteria
        args : Namespace
            Namespace containing all user given arguments compiled by arg_parse()
        **kwargs
            extra parameters passed onto super class ``multiprocessing.Queue()``
        """
        Process.__init__(self, **kwargs)
        self.input_queue = input_queue
        self.counter_queue = counter_queue
        self.writer_queue = writer_queue
        self.vcf_file_obj = VCF(args.vcf)
        # header used for converting bam string to bam object
        self.header = pysam.AlignmentFile(args.bam, 'r').header
        self.args = args

    def run(self):
        """
        Start the ``Processor``

        ``run()`` is automatically called when the scheduler runs ``Processor.start()``,
        this begins the operations of both the ``multiprocessing.Process()`` super
        class and ``Processor``.

        See ``check_variants()``, ``cigar()``, and ``phase_detection`` for
        more information on the analysis process.
        """
        pair = self.input_queue.get(block=True)

        while pair:
            # use bam header to convert bam string back to pysam bam object
            record_1 = pysam.AlignedSegment.fromstring(pair[0], self.header)
            record_2 = pysam.AlignedSegment.fromstring(pair[1], self.header)

            # skip if either record is below mapq filter
            if any([record_1.mapq < self.args.min_mapq, record_2.mapq < self.args.min_mapq]):
                pair = self.input_queue.get(block=True)
                self.counter_queue.put('seq')
                self.counter_queue.put('failed_mapq')
                continue

            variants_1 = check_variants(
                self.vcf_file_obj, record_1.reference_name, record_1.reference_start,
                record_1.reference_start + record_1.query_alignment_length)
            variants_2 = check_variants(
                self.vcf_file_obj, record_2.reference_name, record_2.reference_start,
                record_2.reference_start + record_2.query_alignment_length)

            self.counter_queue.put('seq')
            if len(variants_1) + len(variants_2) > 1:
                self.counter_queue.put('seq_with_variants')
            # skip if not enough variants
            else:
                pair = self.input_queue.get(block=True)
                continue

            segment_1 = cigar(record_1)
            segment_2 = cigar(record_2)

            variant_lst = phase_detection(variants_1, segment_1, record_1, self.args)
            variant_lst += phase_detection(variants_2, segment_2, record_2, self.args)

            if '1' in variant_lst and '2' in variant_lst:
                self.counter_queue.put('phase_change')

                if 'phase_change' in self.args.mode:
                    self.writer_queue.put(pair)

            if 'N' in variant_lst:
                self.counter_queue.put('no_match')

                if 'no_match' in self.args.mode:
                    self.writer_queue.put(pair)

            pair = self.input_queue.get(block=True)

class Counter(Process):
    def __init__(self, input_queue, args, **kwargs):
        """
        Intializes the ``Counter`` process.

        Inherited from ``multiprocessing.Process``, ``Counter`` receives sequence information
        from Processor processes through ``input_queue`` and tallies them to generate a ``tqdm``
        progress bar in ``STDOUT``.

        After being passed a ``None`` from the ``Scheduler, ``Counter``
        either writes tallied results to ``STDOUT`` or a log depending on the user's arguments.

        Parameters
        ----------
        input_queue : multiprocessing.Queue
            get sequence information from all ``Processor`` processes
        args : Namespace
            Namespace containing all user given arguments compiled by arg_parse()
        **kwargs
            extra args that are passed onto super class ``Process``
        """
        Process.__init__(self, **kwargs)
        self.input_queue = input_queue
        self.counters = {
            'no_match': 0,
            'phase_change': 0,
            'seq_with_variants': 0,
            'seq': 0,
            'failed_mapq': 0
            }
        self.args = args
        self.progress = None

    def run(self):
        """
        Start the ``Counter``

        ``run()`` is automatically called when the scheduler runs ``Counter.start()``,
        this begins the operations of both the ``multiprocessing.Process()`` super class and
        ``Counter``
        """

        # start timer
        start = time.time()

        # create tqdm iteration counter if no bars
        if self.args.pair_count:
            self.progress = tqdm(total=self.args.pair_count)
        else:
            self.progress = tqdm()

        count = self.input_queue.get(block=True)

        while count:
            self.counters[count] += 1
            # update iteration counter if sequence
            if count == 'seq':
                self.progress.update(n=1)

            count = self.input_queue.get(block=True)

        self.progress.close()

        # end timer
        end = time.time()
        runtime = str(datetime.timedelta(seconds=round(end - start)))

        # print counters to STDOUT
        print(f'[readcomb] completed filtering for {self.args.bam}')
        print('{} phase change-containing read pairs from total {} read pairs'.format(
            self.counters['phase_change'], self.counters['seq']))
        print('{} reads had no-match variants'.format(self.counters['no_match']))
        print('{} reads did not have enough variants (> 0) to call'.format(
            self.counters['seq'] - self.counters['seq_with_variants']
        ))
        print('{} reads were below the MAPQ threshold ({})'.format(
            self.counters['failed_mapq'], self.args.min_mapq))
        print(f'time taken: {runtime}'.format(runtime))

        # write to log if argument defined
        if self.args.log:
            # determine if we are adding to file or creating new log file
            needs_header = bool(not os.path.isfile(self.args.log))

            # convert argparse namespace to dictionary to be more easily manipulated
            args_dict = vars(self.args)

            with open(self.args.log, 'a') as f:
                if needs_header:
                    fieldnames = ['time',
                                'phase_change_reads',
                                'no_match_reads',
                                'seq_with_variants',
                                'no_variant_reads',
                                'failed_mapq_reads',
                                'seqs',
                                'time_taken']
                    fieldnames.extend(sorted(args_dict.keys()))
                    f.write(','.join(fieldnames) + '\n')

                out_values = [datetime.datetime.now(),
                            self.counters['phase_change'],
                            self.counters['no_match'],
                            self.counters['seq_with_variants'],
                            self.counters['seq'] - self.counters['seq_with_variants'],
                            self.counters['failed_mapq'],
                            self.counters['seq'],
                            runtime]
                out_values.extend([args_dict[key] for key in sorted(args_dict.keys())])
                f.write(','.join([str(n) for n in out_values]) + '\n')

class Writer(Process):
    def __init__(self, input_queue, args, **kwargs):
        """
        Intializes the ``Writer`` process.

        Inherited from ``multiprocessing.Process``, ``Writer`` receives bam records
        in string form from ``Processor`` processes that fulfill the filters given by the user.
        ``Writer`` converts bam records in string form to ``pysam.AlignedSegment`` objects and
        writes them to a SAM file. When ``Writer`` receives a ``None`` from the Scheduler, the
        writing loop is ended and the output file is closed.


        Parameters
        ----------
        input_queue : multiprocessing.Queue
            queue to receive bam record strings from all ``Processor`` processes
        args : Namespace
            Namespace containing all user given arguments compiled by arg_parse()
        **kwargs
            extra args that are passed onto super class ``Process``
        """
        Process.__init__(self, **kwargs)
        self.input_queue = input_queue

        pysam_obj = pysam.AlignmentFile(args.bam, 'r')
        if args.out.endswith('.bam'):
            print('[readcomb] WARNING: readcomb-filter only outputs .sam files')
            args.out = args.out.replace('.bam', '.sam')
            print(f'[readcomb] output will be written to {args.out}')
        elif not args.out.endswith('.sam'):
            args.out = args.out + '.sam'
        self.out = pysam.AlignmentFile(
            args.out, 'wh', template=pysam_obj)
        self.header = pysam.AlignmentFile(args.bam, 'r').header

    def run(self):
        """
        Start the ``Writer``

        ``run()`` is automatically called when the scheduler runs ``Writer.start()``,
        this begins the operations of both the ``multiprocessing.Process()`` super class and
        ``Writer``
        """
        pair = self.input_queue.get(block=True)
        while pair:
            for str_record in pair:
                record = pysam.AlignedSegment.fromstring(str_record, self.header)
                self.out.write(record)

            pair = self.input_queue.get(block=True)
        # very important to close or last piece of information
        # in buffer will not write to file
        self.out.close()


def matepair_process():
    """
    Start processes, queues, and distribute bam records to ``Processors``

    ``matepair_process`` calls ``arg_parse()`` to parse user given arguments.

    It then initializes and starts ``Processors``, ``Counter``, and ``Writer``
    processes/daemons and then equally distribute the bam records
    in string form to ``Processors``.

    After all bam records have been distributed, ``None`` is passed to ``Processor``
    processes, ``Counter process, and ``Writer`` process to signal their stop and
    ``close()`` the processes.
    """
    parser = arg_parser()
    args = parser.parse_args()

    # idxstats on bam/bai file
    if not os.path.isfile(args.bam.replace('.bam', '.bai')):
        print('[readcomb] bai file not provided - continuing without progress bars')
        args.pair_count = None
    else:
        stats = subprocess.check_output(['samtools', 'idxstats', args.bam])
        stats_table = pd.read_csv(
            BytesIO(stats), sep='\t',
            names=['chrom', 'length', 'map_reads', 'unmap_reads'])
        pairs_sum = stats_table['map_reads'].sum()
        args.pair_count = int(pairs_sum / 2) # divide 2 to get pairs

        if pairs_sum % 2 != 0:
            raise ValueError('Preprocessing of bam file went wrong')

    # pysam argument object
    bam = pysam.AlignmentFile(args.bam, 'r')

    print('[readcomb] creating processes')
    # set up counter and writer
    count_input = Queue()
    # daemon causes other sub-processes to auto shutdown if main process is interrupted
    counter = Counter(count_input, args, daemon=True)
    counter.start()

    write_input = Queue()
    writer = Writer(write_input, args, daemon=True)
    writer.start()

    processes = []
    input_queues = []
    # set up processes
    for i in range(args.processes):
        input_queue = Queue()
        input_queues.append(input_queue)
        processes.append(
            Processor(input_queue, count_input, write_input, args,
                daemon=True, name='Process ' + str(i)))
        processes[i].start()

    prev_record = None
    process_idx = 0

    print('[readcomb] processes created')

    for record in bam:
        # check if record is in a proper pair
        if not record.is_proper_pair or record.is_secondary or record.is_supplementary:
            raise ValueError('Preprocessing of bam file went wrong')

        if not prev_record:
            prev_record = record
        else:
            if prev_record.query_name != record.query_name:
                raise ValueError(f'Read {record.query_name} not paired')

            # convert bam sequence to string so it can be pickled and sent in queue
            pair = [prev_record.to_string(), record.to_string()]

            input_queues[process_idx].put(pair)

            # give record to next process
            if process_idx < args.processes - 1:
                process_idx += 1
            else:
                process_idx = 0

            # clear the pair
            prev_record = None

    # shut down processes
    for i, _ in enumerate(processes):
        # pass None to let processes know all bam sequences have been scheduled
        input_queues[i].put(None)
        input_queues[i].close()
        processes[i].join()

    # shut down counter and writer
    count_input.put(None)
    count_input.close()
    counter.join()

    write_input.put(None)
    write_input.close()
    writer.join()

if __name__ == '__main__':
    matepair_process()



#!/usr/bin/env python3
"""
vcfprep.py - prepare parental VCF for phase change detection and apply further
filters if needed
"""

import subprocess
import argparse
import sys
import numpy as np
from cyvcf2 import VCF
from cyvcf2 import Writer
from tqdm import tqdm

class ReadcombParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write(f'error: {message}\n')
        self.print_help()
        sys.exit(1)

def arg_parser():
    parser = ReadcombParser(
        description='prep parental VCF for phase change detection',
        usage='readcomb-vcfprep [options]')

    parser.add_argument('-v', '--vcf', required=True,
        type=str, help='File to filter (.vcf.gz)')
    parser.add_argument('--snps_only', required=False,
        action='store_true', help='Keep only SNPs [optional]')
    parser.add_argument('--indels_only', required=False,
        action='store_true', help='Keep only indels [optional]')
    parser.add_argument('--no_hets', required=False,
        action='store_true', help='Remove heterozygote calls [optional]')
    parser.add_argument('--min_GQ', required=False, default=30,
        type=int, help='Min GQ at both sites (default 30)')
    parser.add_argument('--purity_filter', required=False, default=1,
        type=int, help='Purity filter stringency (default 1 read allowed) - set -1 to disable')
    parser.add_argument('-o', '--out', required=True,
        type=str, help='File to write to. If .gz, script will bgzip and tabix file.')
    parser.add_argument('--version', action='version', version='readcomb 0.3.13')

    return parser

def vcfprep(args):
    """
    Iterate through parental VCF and remove records by given filters.

    All records where parent1 allele is the same as parent2
    ('uninformative sites') are automatically removed.

    Raises ValueError if >2 samples in VCF (should just be 2 parents)

    Parameters
    ----------
    args : Namespace
        Namespace containing all user given arguments compiled by arg_parse()

    Returns
    -------
    total_count : int
        total records considered
    kept_count : int
        number of records that passed filters
    """
    vcf_in = VCF(args.vcf)

    if len(vcf_in.samples) > 2:
        raise ValueError('more than 2 parental samples in input VCF')
    if args.snps_only and args.indels_only:
        raise ValueError('both --snps_only and --indels_only provided. pick one or neither!')

    if args.out.endswith('.gz'):
        outfile = args.out.replace('.gz', '')
    else:
        outfile = args.out
    vcf_out = Writer(outfile, vcf_in)
    vcf_out.write_header()

    total_count, kept_count = 0, 0

    for record in tqdm(vcf_in):
        total_count += 1

        if not len(record.ALT) > 0:
            continue

        # SNP filter
        if args.snps_only and not record.is_snp:
            continue

        # indel filter
        if args.indels_only and not record.is_indel:
            continue

        # heterozygote call filter
        if args.no_hets and record.num_het != 0:
            continue

        # genotype quality filter
        if not all(record.gt_quals >= args.min_GQ):
            continue

        # ensure parental alleles differ
        if record.gt_bases[0][0] == record.gt_bases[1][0]:
            continue

        # purity filter
        purity_threshold = args.purity_filter
        if not purity_threshold == -1:
            # allow at most [threshold] read with opposite parent allele by default
            if len(record.ALT) == 1:
                record_depths = np.array([d[0] for d in record.format('DP')], dtype='int')
                if min(record_depths - record.gt_alt_depths) > purity_threshold:
                    continue
                if min(record_depths - record.gt_ref_depths) > purity_threshold:
                    continue
            elif len(record.ALT) > 1:
                # need to get allele-specific depths from raw record - obtained from AO field
                depths = [call.split(':')[-2].split(',') for call in record.__str__().split('\t')[-2:]]
                # only keep depths for non-ref alleles
                call1_depths, call2_depths = [np.array(d, dtype='int') for d in depths]
                # depth for opposite alt allele should be no more than [threshold]
                if min(call1_depths) > purity_threshold:
                    continue
                if min(call2_depths) > purity_threshold:
                    continue

        # remove any calls with deleted alleles
        if '*' in ' '.join(record.gt_bases):
            continue

        # only passes if record not caught in above filters
        vcf_out.write_record(record)
        kept_count += 1

    return total_count, kept_count

def main():
    parser = arg_parser()
    args = parser.parse_args()

    print(f'[readcomb] Filtering {args.vcf}')
    if args.min_GQ < 30:
        print('[readcomb] WARNING: min GQ below 30 selected')
    if not args.purity_filter == -1:
        print(f'[readcomb] Purity threshold is {args.purity_filter}')
        print('[readcomb] Set --purity_filter to -1 to disable purity filtering')
    total_count, kept_count = vcfprep(args)
    print('[readcomb] Complete.')
    print(f'[readcomb] {kept_count} of {total_count} records retained.')
    if args.out.endswith('.gz'):
        print('[readcomb] compressing outfile...')
        proc_bgzip = subprocess.Popen(['bgzip', args.out.replace('.gz', '')])
        stdout, stderr = proc_bgzip.communicate()
        print('[readcomb] creating tabix index...')
        proc_tabix = subprocess.Popen(['tabix', args.out])
        stdout, stderr = proc_tabix.communicate()
        print('[readcomb] done.')

if __name__ == '__main__':
    main()
